"""
This module implements rules. Both the structure of rules as well as
actual built-in rules are defined.
"""

from dataclasses import dataclass
from enum import Enum
from fnmatch import fnmatch
import re
from typing import Union, Optional, List


class RuleTypes(str, Enum):
    """Allowed types/choices for Rules"""

    REGEX = "regex"


class Rule:
    """Parent class for rules
    Attributes:
        type (RuleTypes): Defines the type of Rule, e.g. Regex
    """

    type: RuleTypes


class RegexRule(Rule):
    """Definition of Regex Rule that applies to a column content
    Attributes:
        name (str): Name of rule
        description (str): Description of the rule
        definition (str): The actual definition. I.e. for regex-type
            rules this would contain the actual regular expressions as
            a RAW string, i.e. r'...'
        match_example (str, optional): It is possible to provide an example of
            a string/expression to be matched by the Rule. This example
            is used to validate the Rule upon instantiation.
        class_name (str, optional): Name of the class used to select columns in unity
            catalog
    """

    def __init__(self, name, description, definition, match_example=[], nomatch_example=[], class_name=None):
        self.name = name
        self.description = description
        self.definition = definition
        self.match_example = match_example
        self.nomatch_example = nomatch_example
        self.class_name = class_name
        self.type = RuleTypes.REGEX

        self.validate_rule(self.match_example, self.definition, self.name, False)
        self.validate_rule(self.nomatch_example, self.definition, self.name, True)

    type: RuleTypes = RuleTypes.REGEX

    name: str
    description: str
    definition: str
    match_example: Optional[Union[str, List[str]]] = None
    nomatch_example: Optional[Union[str, List[str]]] = None
    class_name: Optional[str] = None

    @staticmethod
    def validate_rule(example, definition, name, fail_match: bool):
        """Validates that given example is matched by defined pattern"""
        if not isinstance(example, list):
            validation_example = [example]
        else:
            validation_example = example

        for ex in validation_example:
            if (not re.match(definition, ex)) != fail_match:
                raise ValueError(f"The definition of the rule {name} does not match the provided example {ex}")
        return example


@dataclass
class RulesList:
    """A list of rules with added properties
    Attributes:
    -----------
    rules (List(Rule), optional): A list of rules, could be either
        built-in or custom-defined. Defaults to None
    """

    rules: Optional[List[Rule]] = None

    @property
    def rules_info(self) -> str:
        """Get a HTML-formatted description of all the rules"""
        if self.rules:
            text = [f"<li>{rule.name} - {rule.description}</li>" for rule in self.rules]
            return "\n              ".join(text)

        return ""

    @property
    def number_of_rules(self) -> int:
        """Return the number of rules in the list."""
        if self.rules:
            return len(self.rules)
        return 0

    def test_match(self, input_string: str):
        """Return a list of rules which match the given input string"""
        if self.rules is not None:
            return [rule.name for rule in self.rules if re.match(rule.definition, input_string)]

        return []


# define builtin rules
global_rules = [
    RegexRule(
        name="credit_card_expiration_date",
        description="Credit Card Expiration Date",
        definition=r"^\d{2}/\d{2}$",
        match_example=["01/20", "12/25"],
        nomatch_example=["1/20", "01/2020", "01/2", "01/200"],
    ),
    RegexRule(
        name="credit_card_number",
        description="Credit Card Number",
        definition=r"^\d{4}-\d{4}-\d{4}-\d{4}$",
        match_example=["1234-5678-9012-3456", "9876-5432-1098-7654"],
        nomatch_example=["1234-5678-9012-345", "1234-5678-9012-34567", "1234-5678-9012-3456-7890"],
    ),
    RegexRule(
        name="decimal_number",
        description="Decimal Number",
        definition=r"^-?\d+(?:[.,]\d*)?[eE]?-?\d{0,3}$",
        match_example=[
            "123.45",
            "-123.45",
            "1.1E2",
            "1.1E-1",
            "123,45",
            "-123,45",
            "123,0123",
            "-123,0",
            "123.",
            "123,",
            "-123.",
            "-123,",
        ],
        nomatch_example=["", "123,456,789", "1$", "123,456.789"],
    ),
    RegexRule(
        name="email",
        description="Email address",
        definition=r"^.+@[^\.].*\.[a-z]{2,}$",
        match_example=[
            "whatever@somewhere.museum",
            "foreignchars@myforeigncharsdomain.nu",
            "me+mysomething@mydomain.com",
        ],
        nomatch_example=["a@b.c", "me@.my.com", "a@b.comFOREIGNCHAR"],
    ),
    RegexRule(
        name="fqdn",
        description="Fully Qualified Domain Names",
        definition=r"^([-a-zA-Z0-9:%._\+~#=]{1,63}\.){1,8}[a-zA-Z]{1,12}\.?$",
        match_example=[
            "ec2-35-160-210-253.us-west-2-.compute.amazonaws.com",
            "ec2-35-160-210-253.us-west-2-.compute.amazonaws.com.mx.gmail.com.",
            "1.2.3.4.com",
            "xn--kxae4bafwg.xn--pxaix.gr",
        ],
        nomatch_example=[
            "so-me.na-me.567",
            "label.name.321",
            "1234567890-1234567890-1234567890-1234567890-12345678901234567890.123.com",
            "abc.cdf@mydoamain.com",
            "Some text abc.cdf.com",
        ],
    ),
    RegexRule(
        name="integer_number",
        description="Integer Number",
        definition=r"^-?\d+$",
        match_example=["123", "-123", "0"],
        nomatch_example=["", "123.45", "123,45", "123,0", "123.0", "123,456,789", "1$"],
    ),
    RegexRule(
        name="ip_v4",
        description="IP address v4",
        definition=r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$",
        match_example=["192.1.1.1", "0.0.0.0"],
        nomatch_example=["192"],
    ),
    RegexRule(
        name="ip_v6",
        description="IP address v6",
        definition=r"^(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))$",
        match_example=["2001:db8:3333:4444:5555:6666:7777:8888", "::1234:5678", "2001:db8::", "::"],
        nomatch_example=["2001.0000"],
    ),
    RegexRule(
        name="iso_date",
        description="ISO Date",
        definition=r"^\d{4}-\d{2}-\d{2}$",
        match_example=["2020-01-01", "2020-12-31"],
        nomatch_example=["2020-01", "2020-01-01-01", "2020-01-01T01:01:01"],
    ),
    RegexRule(
        name="iso_date_time",
        description="ISO Date Time",
        definition=r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}$",
        match_example=["2020-01-01T01:01:01", "2020-12-31T23:59:59"],
        nomatch_example=["2020-01", "2020-01-01", "2020-01-01-01"],
    ),
    RegexRule(
        name="mac_address",
        description="MAC Addresses",
        definition=r"^(?=[-:\w]*[a-fA-F]+[-:\w]*)(([0-9A-Fa-f]{2}[:-]?){5}([0-9A-Fa-f]{2}))$",
        match_example=["01:02:03:04:ab:cd", "01-02-03-04-ab-cd", "0102-0304-abcd", "01020304abcd"],
        nomatch_example=[
            "01:02:03:04:ab",
            "01.02.03.04.ab.cd",
            "01:02:03:04:05:06",  # There must be at least one letter
        ],
    ),
    RegexRule(
        name="url",
        description="URL",
        definition=r"^(https?|ftp|file|mailto):\/\/(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b(?:[-a-zA-Z0-9()@:%_\+.~#?&\/=]*)$",
        match_example=[
            "http://www.domain.com",
            "http://domain.com",
            "http://domain.com",
            "https://domain.com",
            "https://sub.domain-name.com:8080",
            "http://domain.com/dir%201/dir_2/program.ext?var1=x&var2=my%20value",
            "ftp://domain.com/index.html#bookmark",
            "file://domain.com/abc.txt",
        ],
        nomatch_example=["Some text http://domain.com", "http://domain.com some text", "my@email.com"],
    ),
]

localized_rules = {
    "us": [
        RegexRule(
            name="us_mailing_address",
            description="US Mailing Address",
            definition=r"^\d+\s[A-z]+\s[A-z]+",
            match_example=["123 Main St", "456 Elm St", "789 Pine St"],
            nomatch_example=["123 Main", "456 Elm", "789 Pine"],
        ),
        RegexRule(
            name="us_phone_number",
            description="US Phone Number",
            definition=r"^\+?1?[-. (]*(\d{3})[-. )]*(\d{3})[-. ]*(\d{4})$",
            match_example=["+1 (123) 456-7890", "123-456-7890", "123.456.7890", "1234567890", "(123)456-7890"],
            nomatch_example=["123-45-6789", "987-65-4321"],
        ),
        RegexRule(
            name="us_social_security_number",
            description="US Social Security Number",
            definition=r"^(?!000|666|9)\d{3}-(?!00)\d{2}-(?!0000)\d{4}$",
            match_example=["123-45-6789"],
            nomatch_example=["123-45-678", "123-456-7890", "123-45-67890", "123-456-789"],
        ),
        RegexRule(
            name="us_state",
            description="US State",
            definition=r"(?i)^(Alabama|Alaska|American Samoa|Arizona|Arkansas|California|Colorado|Connecticut|Delaware|District of Columbia|Federated States of Micronesia|Florida|Georgia|Guam|Hawaii|Idaho|Illinois|Indiana|Iowa|Kansas|Kentucky|Louisiana|Maine|Marshall Islands|Maryland|Massachusetts|Michigan|Minnesota|Mississippi|Missouri|Montana|Nebraska|Nevada|New Hampshire|New Jersey|New Mexico|New York|North Carolina|North Dakota|Northern Mariana Islands|Ohio|Oklahoma|Oregon|Palau|Pennsylvania|Puerto Rico|Rhode Island|South Carolina|South Dakota|Tennessee|Texas|Utah|Vermont|Virgin Islands|Virginia|Washington|West Virginia|Wisconsin|Wyoming)$",
            match_example=[
                "Alabama",
                "Alaska",
                "Arizona",
                "Arkansas",
                "California",
                "Colorado",
                "Connecticut",
                "Delaware",
                "Florida",
                "Georgia",
                "Hawaii",
                "Idaho",
                "Illinois",
                "Indiana",
                "Iowa",
                "Kansas",
                "Kentucky",
                "Louisiana",
                "Maine",
                "Maryland",
                "Massachusetts",
                "Michigan",
                "Minnesota",
                "Mississippi",
                "Missouri",
                "Montana",
                "Nebraska",
                "Nevada",
                "New Hampshire",
                "New Jersey",
                "New Mexico",
                "New York",
                "North Carolina",
                "North Dakota",
                "Ohio",
                "Oklahoma",
                "Oregon",
                "Pennsylvania",
                "Rhode Island",
                "South Carolina",
                "South Dakota",
                "Tennessee",
                "Texas",
                "Utah",
                "Vermont",
                "Virginia",
                "Washington",
                "West Virginia",
                "Wisconsin",
                "Wyoming",
            ],
            nomatch_example=[
                "AL",
                "AK",
                "AS",
                "AZ",
                "AR",
                "CA",
                "CO",
                "CT",
                "DE",
                "DC",
                "FM",
                "FL",
                "GA",
                "GU",
                "HI",
                "ID",
                "IL",
                "IN",
                "IA",
                "KS",
                "KY",
                "LA",
                "ME",
                "MH",
                "MD",
                "MA",
                "MI",
                "MN",
                "MS",
                "MO",
                "MT",
                "NE",
                "NV",
                "NH",
                "NJ",
                "NM",
                "NY",
                "NC",
                "ND",
                "MP",
                "OH",
                "OK",
                "OR",
                "PW",
                "PA",
                "PR",
                "RI",
                "SC",
                "SD",
                "TN",
                "TX",
                "UT",
                "VT",
                "VI",
                "VA",
                "WA",
                "WV",
                "WI",
                "WY",
            ],
        ),
        RegexRule(
            name="us_state_abbreviation",
            description="US State Abbreviation",
            definition=r"(?i)^(AL|AK|AS|AZ|AR|CA|CO|CT|DE|DC|FM|FL|GA|GU|HI|ID|IL|IN|IA|KS|KY|LA|ME|MH|MD|MA|MI|MN|MS|MO|MT|NE|NV|NH|NJ|NM|NY|NC|ND|MP|OH|OK|OR|PW|PA|PR|RI|SC|SD|TN|TX|UT|VT|VI|VA|WA|WV|WI|WY)$",
            match_example=[
                "AL",
                "AK",
                "AS",
                "AZ",
                "AR",
                "CA",
                "CO",
                "CT",
                "DE",
                "DC",
                "FM",
                "FL",
                "GA",
                "GU",
                "HI",
                "ID",
                "IL",
                "IN",
                "IA",
                "KS",
                "KY",
                "LA",
                "ME",
                "MH",
                "MD",
                "MA",
                "MI",
                "MN",
                "MS",
                "MO",
                "MT",
                "NE",
                "NV",
                "NH",
                "NJ",
                "NM",
                "NY",
                "NC",
                "ND",
                "MP",
                "OH",
                "OK",
                "OR",
                "PW",
                "PA",
                "PR",
                "RI",
                "SC",
                "SD",
                "TN",
                "TX",
                "UT",
                "VT",
                "VI",
                "VA",
                "WA",
                "WV",
                "WI",
                "WY",
            ],
            nomatch_example=[
                "XY",
                "Alabama",
                "Alaska",
                "Arizona",
                "Arkansas",
                "California",
                "Colorado",
                "Connecticut",
                "Delaware",
                "Florida",
                "Georgia",
                "Hawaii",
                "Idaho",
                "Illinois",
                "Indiana",
                "Iowa",
                "Kansas",
                "Kentucky",
                "Louisiana",
                "Maine",
                "Maryland",
                "Massachusetts",
                "Michigan",
                "Minnesota",
                "Mississippi",
                "Missouri",
                "Montana",
                "Nebraska",
                "Nevada",
                "New Hampshire",
                "New Jersey",
                "New Mexico",
                "New York",
                "North Carolina",
                "North Dakota",
                "Ohio",
                "Oklahoma",
                "Oregon",
                "Pennsylvania",
                "Rhode Island",
                "South Carolina",
                "South Dakota",
                "Tennessee",
                "Texas",
                "Utah",
                "Vermont",
                "Virginia",
                "Washington",
                "West Virginia",
                "Wisconsin",
                "Wyoming",
            ],
        ),
        RegexRule(
            name="us_zip_code",
            description="US Zip Code",
            definition=r"^\d{5}(?:[-\s]\d{4})?$",
            match_example=["12345", "12345-6789"],
            nomatch_example=["1234", "123456"],
        ),
    ]
}


class Rules:
    """This class represents built-in and custom rules

    Attributes:
        builtin_rules (RulesList): Represents all built-in rules
        custom_rules (RulesList): Represents all custom-defined rules

    """

    def __init__(self, locale: Optional[str] = None, custom_rules: Optional[List[Rule]] = None):
        """
        Args:
            locale (str, optional): The locale to use. Defaults to None.
            custom_rules (List[Rule], optional): A list of
                custom-defined rules. Defaults to None.
        """
        rules = global_rules.copy()
        if locale is not None:
            if locale.lower() not in localized_rules:
                raise ValueError(f"Unsupported locale: {locale}. Use one of {localized_rules.keys()}")
            rules += localized_rules[locale.lower()]
        self.builtin_rules: RulesList = RulesList(rules)
        self.custom_rules = RulesList(custom_rules)

    def get_rules_info(self):
        """Returns info about the available rules
        Both built-in and custom rules will be presented

        Returns:
            HTML-formatted string
        """
        text = f"""
                <h2>Matching rules</h2>
                <p>
                  Here are the {self.builtin_rules.number_of_rules} built-in rules that are available to you:
                </p>
                <ul>
                  {self.builtin_rules.rules_info}
                </ul>
                <p>
                  You've also defined {self.custom_rules.number_of_rules} custom rule(s) which is/are:
                </p>
                <ul>
                  {self.custom_rules.rules_info}
                </ul>
                """
        return text

    def get_rules(self, rule_filter: str) -> List[Rule]:
        """Return a list of specified rules
        The rules to be returned are specified using unix wildcards.
        Both built-in and custom rules are returned in one single list.

        Args:
            rule_filter (str): Unix wildcard string specifying the
                rule(s) to be returned. E.g. '*' for all rules while
                '*v4' would return all rules with names matching the
                wildcard. This would include the 'ip_v4' built-in rule.

        Returns:
            A list of rules matching the wildcard
        """
        filtered_builtin_rules = self._filter_rules(self.builtin_rules.rules, rule_filter)
        filtered_custom_rules = self._filter_rules(self.custom_rules.rules, rule_filter)
        return filtered_builtin_rules + filtered_custom_rules

    @staticmethod
    def _filter_rules(rule_set: List[Rule], rule_filter: str) -> List[Rule]:
        """Filter a given set of rules

        Args:
            rule_set (List[Rule]): The set of rules to be filtered, e.g.
                the built-in or custom rules.
            rule_filter (str): Unix wildcard string specifying the
                rule(s) to be returned. E.g. '*' for all rules while
                '*v4' would return all rules with names matching the
                wildcard. This would include the 'ip_v4' built-in rule.

        Returns:
            A list of rules from the specified rule set matching the
            wildcard.
        """
        return [rule for rule in (rule_set if rule_set is not None else []) if fnmatch(rule.name, rule_filter)]

    def match_search_term(self, search_term: str):
        """Return list of rules matching the search term

        Args:
            search_term: A string to be matched by the available rules

        Returns:
            A list of matching rules, both custom and built-in
        """
        return self.builtin_rules.test_match(search_term) + self.custom_rules.test_match(search_term)



"""This module contains the M-SQL compiler"""
from dataclasses import dataclass
from functools import reduce
from discoverx import logging
from discoverx.scanner import ColumnInfo, TableInfo
from discoverx.common.helper import strip_margin
from fnmatch import fnmatch
from pyspark.sql.functions import lit
from pyspark.sql import DataFrame, SparkSession
import re
import itertools


@dataclass
class SQLRow:
    catalog: str
    schema: str
    table: str
    sql: str


class Msql:
    """This class compiles M-SQL expressions into regular SQL"""

    from_statement_expr = r"(FROM\s+)(([0-9a-zA-Z_\*]+).([0-9a-zA-Z_\*]+).([0-9a-zA-Z_\*]+))"
    from_components_expr = r"^(([0-9a-zA-Z_\*]+).([0-9a-zA-Z_\*]+).([0-9a-zA-Z_\*]+))$"
    command_expr = r"^\s*(\w+)\s"
    class_regex = r"\[([\w_-]+)\]"
    valid_commands = ["SELECT", "DELETE"]

    def __init__(self, msql: str) -> None:
        self.msql = msql

        # Find distinct classes in M-SQL expression
        self.classes = list(set(re.findall(self.class_regex, msql)))

        # Extract from clause components
        (self.catalogs, self.schemas, self.tables) = self._extract_from_components()

        # Extract command
        self.command = self._extract_command()

        self.logger = logging.Logging()

    def compile_msql(self, table_info: TableInfo) -> list[SQLRow]:
        """
        Compiles the M-SQL (Multiplex-SQL) expression into regular SQL
        Args:
            table_info (TableInfo): Table information

        Returns:
            list[string]: A list of SQL expressions which multiplexes the MSQL expression
        """

        # Replace from clause with table name
        msql = strip_margin(self.msql)
        msql = self._replace_from_statement(msql, table_info)

        # Get all columns matching the classes
        columns_by_class = [table_info.get_columns_by_class(class_name) for class_name in self.classes]

        # Create all possible combinations of classified columns to be queried
        col_class_combinations = list(itertools.product(*columns_by_class))

        # Replace classes in M-SQL expression with column names
        sql_statements = []
        for classified_cols in col_class_combinations:
            temp_sql = msql
            for classified_col in classified_cols:
                temp_sql = temp_sql.replace(f"[{classified_col.class_name}]", classified_col.name)
            sql_statements.append(SQLRow(table_info.catalog, table_info.schema, table_info.table, temp_sql))

        return sql_statements

    def build(self, classified_result_pdf) -> list[SQLRow]:
        """Builds the M-SQL expression into a SQL expression"""

        classified_cols = classified_result_pdf.copy()
        classified_cols = classified_cols[classified_cols["class_name"].isin(self.classes)]
        classified_cols = (
            classified_cols.groupby(["table_catalog", "table_schema", "table_name", "column_name"])
            .aggregate(lambda x: list(x))[["class_name"]]
            .reset_index()
        )

        classified_cols["col_classes"] = classified_cols[["column_name", "class_name"]].apply(tuple, axis=1)
        df = (
            classified_cols.groupby(["table_catalog", "table_schema", "table_name"])
            .aggregate(lambda x: list(x))[["col_classes"]]
            .reset_index()
        )

        # Filter tables by matching filter
        filtered_tables = [
            TableInfo(
                row[0],
                row[1],
                row[2],
                [ColumnInfo(col[0], "", None, col[1]) for col in row[3]],  # col name  # TODO  # TODO  # Classes
            )
            for _, row in df.iterrows()
            if fnmatch(row[0], self.catalogs) and fnmatch(row[1], self.schemas) and fnmatch(row[2], self.tables)
        ]

        if len(filtered_tables) == 0:
            raise ValueError(f"No tables found matching filter: {self.catalogs}.{self.schemas}.{self.tables}")

        sqls = flat_map(self.compile_msql, filtered_tables)

        return sqls

    def execute_sql_row(self, sql_row: SQLRow, spark: SparkSession) -> DataFrame:
        """Executes the SQL statement"""
        try:
            result = (
                spark.sql(sql_row.sql)
                .withColumn("table_catalog", lit(sql_row.catalog))
                .withColumn("table_schema", lit(sql_row.schema))
                .withColumn("table_name", lit(sql_row.table))
            )
            if self.command == "DELETE":
                result = result.withColumn("sql", lit(sql_row.sql))

            table_info_cols = ["table_catalog", "table_schema", "table_name"]
            select_cols = table_info_cols + [col for col in result.columns if col not in table_info_cols]
            result = result.select(*select_cols)

        except Exception as e:
            self.logger.info(f"Unable to execute SQL for {sql_row.catalog}.{sql_row.schema}.{sql_row.table}: {e}")
            result = None

        return result

    def execute_sql_rows(self, sqls: list[SQLRow], spark: SparkSession):
        """Executes the SQL statements"""
        results = [self.execute_sql_row(sql_row, spark) for sql_row in sqls]
        success_results = [result for result in results if result is not None]

        if len(success_results) == 0:
            raise ValueError(f"No SQL statements were successfully executed.")

        return reduce(lambda x, y: x.union(y), success_results)

    def _replace_from_statement(self, msql: str, table_info: TableInfo):
        """Replaces the FROM statement in the M-SQL expression with the specified table name"""
        if table_info.catalog and table_info.catalog != "None":
            replace_with = f"FROM {table_info.catalog}.{table_info.schema}.{table_info.table}"
        else:
            replace_with = f"FROM {table_info.schema}.{table_info.table}"

        return re.sub(self.from_statement_expr, replace_with, msql)

    def _extract_from_components(self):
        """Extracts the catalog, schema and table name from the FROM statement in the M-SQL expression"""
        matches = re.findall(self.from_statement_expr, self.msql)
        if len(matches) > 1:
            raise ValueError(f"Multiple FROM statements found in M-SQL expression: {self.msql}")
        elif len(matches) == 1:
            return (matches[0][2], matches[0][3], matches[0][4])
        else:
            raise ValueError(f"Could not extract table name from M-SQL expression: {self.msql}")

    @staticmethod
    def validate_from_components(from_tables: str):
        """Extracts the catalog, schema and table name from the from_table string"""
        matches = re.findall(Msql.from_components_expr, from_tables)
        if len(matches) == 1 and len(matches[0]) == 4:
            return (matches[0][1], matches[0][2], matches[0][3])
        else:
            raise ValueError(
                f"Invalid from_tables statement '{from_tables}'. Should be a string in format 'table_catalog.table_schema.table_name'. You can use '*' as wildcard."
            )

    def _extract_command(self):
        """Extracts the command from the M-SQL expression"""
        commands = re.findall(self.command_expr, self.msql)
        if len(commands) != 1:
            raise ValueError(
                f"Could not extract command from M-SQL expression: {self.msql}. Valid commands are SELECT and DELETE."
            )

        command = commands[0].upper()
        if command not in self.valid_commands:
            raise ValueError(f"Invalid command: {command}. Valid commands are SELECT and DELETE.")

        return command


def flat_map(f, xs):
    ys = []
    for x in xs:
        ys.extend(f(x))
    return ys



from dataclasses import dataclass
from delta.tables import DeltaTable
import pandas as pd
import concurrent.futures
from pyspark.sql import SparkSession
import pyspark.sql.functions as func
from typing import Optional, List, Set
from pyspark.sql.utils import AnalysisException

from discoverx.common.helper import strip_margin, format_regex
from discoverx import logging
from discoverx.rules import Rules, RuleTypes

logger = logging.Logging()


@dataclass
class ColumnInfo:
    name: str
    data_type: str
    partition_index: int
    classes: list[str]


@dataclass
class TableInfo:
    catalog: Optional[str]
    schema: str
    table: str
    columns: list[ColumnInfo]

    def get_columns_by_class(self, class_name: str):
        return [ClassifiedColumn(col.name, class_name) for col in self.columns if class_name in col.classes]


@dataclass
class ClassifiedColumn:
    name: str
    class_name: str


@dataclass
class ScanContent:
    table_list: List[TableInfo]
    catalogs: Set[str]
    schemas: Set[str]

    @property
    def n_catalogs(self) -> int:
        return len(self.catalogs)

    @property
    def n_schemas(self) -> int:
        return len(self.schemas)

    @property
    def n_tables(self) -> int:
        return len(self.table_list)


@dataclass
class ScanResult:
    df: pd.DataFrame
    spark: SparkSession

    @property
    def is_empty(self) -> bool:
        return self.df.empty

    @property
    def n_scanned_columns(self) -> int:
        return ScanResult.count_distinct_cols(self.df)

    @staticmethod
    def count_distinct_cols(df: pd.DataFrame) -> int:
        return len(df[["table_catalog", "table_schema", "table_name", "column_name"]].drop_duplicates())

    def n_classified_columns(self, min_score: Optional[float]) -> int:
        return ScanResult.count_distinct_cols(self.get_classes(min_score))

    def get_classes(self, min_score: Optional[float]):
        if self.df is None or self.df.empty:
            raise Exception("No scan result available. Please run dx.scan() or dx.load() first.")

        if min_score is None:
            return self.df[self.df["score"] > 0]
        elif (min_score >= 0) and (min_score <= 1):
            return self.df[self.df["score"] >= min_score]
        else:
            error_msg = f"min_score has to be either None or in interval [0,1]. Given value is {min_score}"
            logger.error(error_msg)
            raise ValueError(error_msg)

    def rule_match_str(self, min_score) -> str:
        rule_match_counts = []
        df_summary = self.get_classes(min_score=min_score).groupby(["class_name"]).agg({"score": "count"})
        df_summary = df_summary.reset_index()  # make sure indexes pair with number of rows
        for _, row in df_summary.iterrows():
            rule_match_counts.append(f"            <li>{row['score']} {row['class_name']} columns</li>")
        return "\n".join(rule_match_counts)

    def _create_databes_if_not_exists(self, scan_table_name: str):
        logger.friendly(f"The scan result table {scan_table_name} does not seem to exist. Trying to create it ...")
        (catalog, schema, table) = scan_table_name.split(".")
        try:
            self.spark.sql(f"DESCRIBE CATALOG {catalog}")
        except AnalysisException:
            self.spark.sql(f"CREATE CATALOG IF NOT EXISTS {catalog}")

        try:
            self.spark.sql(f"DESCRIBE DATABASE {catalog + '.' + schema}")
        except AnalysisException:
            self.spark.sql(f"CREATE DATABASE IF NOT EXISTS {catalog + '.' + schema}")

        self.spark.sql(
            f"CREATE TABLE IF NOT EXISTS {scan_table_name} (table_catalog string, table_schema string, table_name string, column_name string, class_name string, score double, effective_timestamp timestamp)"
        )
        logger.friendly(f"The scan result table {scan_table_name} has been created.")

    def _get_or_create_result_table_from_delta(self, scan_table_name: str) -> DeltaTable:
        try:
            return DeltaTable.forName(self.spark, scan_table_name)
        except Exception:
            self._create_databes_if_not_exists(scan_table_name)
            return DeltaTable.forName(self.spark, scan_table_name)

    def save(self, scan_table_name: str):
        scan_delta_table = self._get_or_create_result_table_from_delta(scan_table_name)

        scan_result_df = self.spark.createDataFrame(
            self.df,
            "table_catalog: string, table_schema: string, table_name: string, column_name: string, class_name: string, score: double",
        ).withColumn("effective_timestamp", func.current_timestamp())

        logger.friendly(f"Merging results into {scan_table_name}")

        scan_delta_table.alias("scan_delta_table").merge(
            scan_result_df.alias("scan_result_df"),
            "scan_delta_table.table_catalog = scan_result_df.table_catalog \
            and scan_delta_table.table_schema = scan_result_df.table_schema \
            and scan_delta_table.table_name = scan_result_df.table_name \
            and scan_delta_table.column_name = scan_result_df.column_name ",
        ).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()

    def load(self, scan_table_name: str):
        try:
            self.df = DeltaTable.forName(self.spark, scan_table_name).toDF().drop("effective_timestamp").toPandas()
        except Exception as e:
            logger.error(f"Error while reading the scan result table {scan_table_name}: {e}")
            raise e


class Scanner:
    def __init__(
        self,
        spark: SparkSession,
        rules: Rules,
        catalogs: str = "*",
        schemas: str = "*",
        tables: str = "*",
        rule_filter: str = "*",
        sample_size: int = 1000,
        what_if: bool = False,
        columns_table_name: str = "",
        max_workers: int = 10,
    ):
        self.spark = spark
        self.rules = rules
        self.catalogs = catalogs
        self.schemas = schemas
        self.tables = tables
        self.rules_filter = rule_filter
        self.sample_size = sample_size
        self.what_if = what_if
        self.columns_table_name = columns_table_name
        self.max_workers = max_workers

        self.content: ScanContent = self._resolve_scan_content()
        self.rule_list = self.rules.get_rules(rule_filter=self.rules_filter)
        self.scan_result: Optional[ScanResult] = None

    def _get_list_of_tables(self) -> List[TableInfo]:
        table_list_sql = self._get_table_list_sql()

        rows = self.spark.sql(table_list_sql).collect()
        filtered_tables = [
            TableInfo(
                row["table_catalog"],
                row["table_schema"],
                row["table_name"],
                [
                    ColumnInfo(col["column_name"], col["data_type"], col["partition_index"], [])
                    for col in row["table_columns"]
                ],
            )
            for row in rows
        ]
        return filtered_tables

    def _get_table_list_sql(self):
        """
        Returns a SQL expression which returns a list of columns matching
        the specified filters

        Returns:
            string: The SQL expression
        """

        catalog_sql = f"""AND regexp_like(table_catalog, "^{self.catalogs.replace("*", ".*")}$")"""
        schema_sql = f"""AND regexp_like(table_schema, "^{self.schemas.replace("*", ".*")}$")"""
        table_sql = f"""AND regexp_like(table_name, "^{self.tables.replace("*", ".*")}$")"""

        sql = f"""
        SELECT 
            table_catalog, 
            table_schema, 
            table_name, 
            collect_list(struct(column_name, data_type, partition_index)) as table_columns
        FROM {self.columns_table_name}
        WHERE 
            table_schema != "information_schema" 
            {catalog_sql if self.catalogs != "*" else ""}
            {schema_sql if self.schemas != "*" else ""}
            {table_sql if self.tables != "*" else ""}
        GROUP BY table_catalog, table_schema, table_name
        """

        return strip_margin(sql)

    def _resolve_scan_content(self) -> ScanContent:
        table_list = self._get_list_of_tables()
        catalogs = set(map(lambda x: x.catalog, table_list))
        schemas = set(map(lambda x: f"{x.catalog}.{x.schema}", table_list))

        return ScanContent(table_list, catalogs, schemas)

    def scan_table(self, table: TableInfo):
        try:
            if self.what_if:
                logger.friendly(f"SQL that would be executed for '{table.catalog}.{table.schema}.{table.table}'")
            else:
                logger.friendly(f"Scanning table '{table.catalog}.{table.schema}.{table.table}'")

            # Build rule matching SQL
            sql = self._rule_matching_sql(table)

            if self.what_if:
                logger.friendly(sql)
            else:
                # Execute SQL and return the result
                return self.spark.sql(sql).toPandas()
        except Exception as e:
            logger.error(f"Error while scanning table '{table.catalog}.{table.schema}.{table.table}': {e}")
            return None

    def scan(self):
        logger.friendly("""Ok, I'm going to scan your lakehouse for data that matches your rules.""")
        text = f"""
                This is what you asked for:

                    catalogs ({self.content.n_catalogs}) = {self.catalogs}
                    schemas ({self.content.n_schemas}) = {self.schemas}
                    tables ({self.content.n_tables}) = {self.tables}
                    rules ({len(self.rule_list)}) = {self.rules_filter}
                    sample_size = {self.sample_size}

                This may take a while, so please be patient. I'll let you know when I'm done.
                ...
                """

        logger.friendly(strip_margin(text))

        logger.debug("Launching lakehouse scanning task\n")

        if len(self.content.table_list) == 0:
            raise Exception("No tables found matching your filters")

        dfs = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            # Submit tasks to the thread pool
            futures = [executor.submit(self.scan_table, table) for table in self.content.table_list]

            # Process completed tasks
            for future in concurrent.futures.as_completed(futures):
                result = future.result()
                if result is not None:
                    dfs.append(result)

        logger.debug("Finished lakehouse scanning task")

        if dfs:
            self.scan_result = ScanResult(df=pd.concat(dfs), spark=self.spark)
            return self.scan_result
        else:
            raise Exception("No tables were scanned successfully.")

    def _rule_matching_sql(self, table_info: TableInfo):
        """
        Given a table and a set of rules this method will return a
        SQL expression which matches the table's columns against rules.
        If executed on the table using SQL the output will contain a
        matching score (probability) for each column and rule.
        Args:
            table_info (TableInfo): Specifies the table to be scanned

        Returns:
            string: The SQL expression

        """

        expressions = [r for r in self.rule_list if r.type == RuleTypes.REGEX]
        cols = [c for c in table_info.columns if c.data_type.lower() == "string"]

        if not cols:
            raise Exception(f"There are no columns of type string to be scanned in {table_info.table}")

        if not expressions:
            raise Exception(f"There are no rules to scan for.")

        catalog_str = f"{table_info.catalog}." if table_info.catalog else ""
        matching_columns = [
            f"INT(regexp_like(value, '{format_regex(r.definition)}')) AS `{r.name}`" for r in expressions
        ]
        matching_string = ",\n                    ".join(matching_columns)

        unpivot_expressions = ", ".join([f"'{r.name}', `{r.name}`" for r in expressions])
        unpivot_columns = ", ".join([f"'{c.name}', `{c.name}`" for c in cols])

        sql = f"""
            SELECT 
                '{table_info.catalog}' as table_catalog,
                '{table_info.schema}' as table_schema,
                '{table_info.table}' as table_name, 
                column_name,
                class_name,
                (sum(value) / count(value)) as score
            FROM
            (
                SELECT column_name, stack({len(expressions)}, {unpivot_expressions}) as (class_name, value)
                FROM 
                (
                    SELECT
                    column_name,
                    {matching_string}
                    FROM (
                        SELECT
                            stack({len(cols)}, {unpivot_columns}) AS (column_name, value)
                        FROM {catalog_str}{table_info.schema}.{table_info.table}
                        TABLESAMPLE ({self.sample_size} ROWS)
                    )
                )
            )
            GROUP BY table_catalog, table_schema, table_name, column_name, class_name
        """

        return strip_margin(sql)

    @property
    def summary_html(self) -> str:
        # Summary
        classified_cols = self.scan_result.get_classes(min_score=None)
        classified_cols.index = pd.MultiIndex.from_frame(
            classified_cols[["table_catalog", "table_schema", "table_name", "column_name"]]
        )
        summary_html_table = classified_cols[["class_name", "score"]].to_html()

        return f"""
        <h2>Result summary</h2>
        <p>
          I've been able to classify {self.scan_result.n_classified_columns(min_score=None)} out of {self.scan_result.n_scanned_columns} columns.
        </p>
        <p>
          I've found:
          <ul>
            {self.scan_result.rule_match_str(min_score=None)}
          </ul>
        </p>
        <p>
          To be more precise:
        </p>
        {summary_html_table}
        <p>
          You can see the full classification output with 'dx.scan_result'.
        </p>


        """



import pandas as pd
from pyspark.sql import SparkSession
from typing import List, Optional, Union
from discoverx import logging
from discoverx.msql import Msql
from discoverx.rules import Rules, Rule
from discoverx.scanner import Scanner, ScanResult


class DX:
    """DiscoverX scans and searches your lakehouse

    DiscoverX scans your data for patterns which have been pre-defined
    as rules. You can either use standard rules which come with
    DiscoverX or define and add custom rules.
    Attributes:
        custom_rules (List[Rule], Optional): Custom rules which will be
            used to detect columns with corresponding patterns in your
            data
        spark (SparkSession, optional): The SparkSession which will be
            used to scan your data. Defaults to None.
        locale (str, optional): The two-letter country code which will be
            used to determine the localized scanning rules.
            Defaults to None.
    """

    COLUMNS_TABLE_NAME = "system.information_schema.columns"
    MAX_WORKERS = 10

    def __init__(
        self,
        custom_rules: Optional[List[Rule]] = None,
        spark: Optional[SparkSession] = None,
        locale: str = None,
    ):
        if spark is None:
            spark = SparkSession.getActiveSession()
        self.spark = spark
        self.logger = logging.Logging()

        self.rules = Rules(custom_rules=custom_rules, locale=locale)
        self.uc_enabled = self.spark.conf.get("spark.databricks.unityCatalog.enabled", "false") == "true"

        self.scanner: Optional[Scanner] = None
        self._scan_result: Optional[ScanResult] = None

        self.intro()

    def _can_read_columns_table(self) -> bool:
        try:
            self.spark.sql(f"SELECT * FROM {self.COLUMNS_TABLE_NAME} LIMIT 1")
            return True
        except Exception as e:
            self.logger.error(f"Error while reading table {self.COLUMNS_TABLE_NAME}: {e}")
            return False

    def intro(self):
        # TODO: Decide on how to do the introduction
        intro_text = """
        <h1>Hi there, I'm DiscoverX.</h1>

        <p>
          I'm here to help you discover data in your lakehouse.<br />
          You can scan your lakehouse by using
        </p>
        <pre><code>dx.scan(from_tables="*.*.*")</code></pre>
        <p>
          For more detailed instructions, check out the <a href="https://github.com/databrickslabs/discoverx">readme</a> or use
        </p>
        <pre><code>help(DX)</code></pre>
        """

        missing_uc_text = """
        <h1 style="color: red">Uch! DiscoverX needs Unity Catalog to be enabled</h1>

        <p>
          Please make sure you have Unity Catalog enabled, and that you are running a Cluster that supports Unity Catalog.
        </p>
        """

        missing_access_to_columns_table_text = """
        <h1 style="color: red">DiscoverX needs access to the system tables `system.information_schema.columns`</h1>
        
        <p>
            Please make sure you have access to the system tables `system.information_schema.columns` and that you are running a Cluster that supports Unity Catalog.
            To grant access to the system tables, execute the following commands:<br />
            <code>GRANT USE CATALOG ON CATALOG system TO `account users`;</code><br /> 
            <code>GRANT USE SCHEMA ON CATALOG system TO `account users`;</code><br />
            <code>GRANT SELECT ON CATALOG system TO `account users`;</code>
        </p>
        """

        if not self.uc_enabled:
            self.logger.friendlyHTML(missing_uc_text)
        if not self._can_read_columns_table():
            self.logger.friendlyHTML(missing_access_to_columns_table_text)
        else:
            self.logger.friendlyHTML(intro_text)

    def display_rules(self):
        """Displays the available rules in a friendly HTML format"""
        text = self.rules.get_rules_info()
        self.logger.friendlyHTML(text)

    def scan(
        self,
        from_tables="*.*.*",
        rules="*",
        sample_size=10000,
        what_if: bool = False,
    ):
        """Scans the lakehouse for columns matching the given rules

        Args:
            from_tables (str, optional): The tables to be scanned in format "catalog.schema.table", use "*" as a wildcard. Defaults to "*.*.*".
            rules (str, optional): The rule names to be used to scan the lakehouse, use "*" as a wildcard. Defaults to "*".
            sample_size (int, optional): The number of rows to be scanned per table. Defaults to 10000.
            what_if (bool, optional): Whether to run the scan in what-if mode and print the SQL commands instead of executing them. Defaults to False.
        """
        catalogs, schemas, tables = Msql.validate_from_components(from_tables)

        self.scanner = Scanner(
            self.spark,
            self.rules,
            catalogs=catalogs,
            schemas=schemas,
            tables=tables,
            rule_filter=rules,
            sample_size=sample_size,
            what_if=what_if,
            columns_table_name=self.COLUMNS_TABLE_NAME,
            max_workers=self.MAX_WORKERS,
        )

        self._scan_result = self.scanner.scan()
        self.logger.friendlyHTML(self.scanner.summary_html)

    def _check_scan_result(self):
        if self._scan_result is None:
            raise Exception("You first need to scan your lakehouse using Scanner.scan()")

    @property
    def scan_result(self):
        """Returns the scan results as a pandas DataFrame

        Raises:
            Exception: If the scan has not been run
        """
        self._check_scan_result()

        return self._scan_result.df

    def save(self, full_table_name: str):
        """Saves the scan results to the lakehouse

        Args:
            full_table_name (str): The full table name to be
                used to save the scan results.
        Raises:
            Exception: If the scan has not been run

        """
        self._check_scan_result()
        # save classes
        self._scan_result.save(full_table_name)

    def load(self, full_table_name: str):
        """Loads previously saved scan results from a table

        Args:
            full_table_name (str, optional): The full table name to be
                used to load the scan results.
        Raises:
            Exception: If the table to be loaded does not exist
        """
        self._scan_result = ScanResult(df=pd.DataFrame(), spark=self.spark)
        self._scan_result.load(full_table_name)

    def search(
        self,
        search_term: str,
        from_tables: str = "*.*.*",
        by_class: Optional[str] = None,
        min_score: Optional[float] = None,
    ):
        """Searches your lakehouse for columns matching the given search term

        Args:
            search_term (str): The search term to be used to search for columns.
            from_tables (str, optional): The tables to be searched in format
                "catalog.schema.table", use "*" as a wildcard. Defaults to "*.*.*".
            by_class (str, optional): The class to be used to search for columns.
                Defaults to None.
            min_score (float, optional): Defines the classification score or frequency
                threshold for columns to be considered during the scan. Defaults to None
                which means that all columns where at least one record matched the
                respective rule during the scan will be included. Has to be either None
                or between 0 and 1.

        Raises:
            ValueError: If search_term is not provided
            ValueError: If the search_term type is not valid
            ValueError: If the by_class type is not valid

        Returns:
            DataFrame: A dataframe containing the results of the search
        """

        Msql.validate_from_components(from_tables)

        if search_term is None:
            raise ValueError("search_term has not been provided.")

        if not isinstance(search_term, str):
            raise ValueError(f"The search_term type {type(search_term)} is not valid. Please use a string type.")

        if by_class is None:
            # Trying to infer the class by the search term
            self.logger.friendly(
                "You did not provide any class to be searched."
                "We will try to auto-detect matching rules for the given search term"
            )
            search_matching_rules = self.rules.match_search_term(search_term)
            if len(search_matching_rules) == 0:
                raise ValueError(
                    f"Could not infer any class for the given search term. Please specify the by_class parameter."
                )
            elif len(search_matching_rules) > 1:
                raise ValueError(
                    f"Multiple classes {search_matching_rules} match the given search term ({search_term}). Please specify the class to search in with the by_class parameter."
                )
            else:
                by_class = search_matching_rules[0]
            self.logger.friendly(f"Discoverx will search your lakehouse using the class {by_class}")
        elif isinstance(by_class, str):
            search_matching_rules = [by_class]
        else:
            raise ValueError(f"The provided by_class {by_class} must be of string type.")

        sql_filter = f"`[{search_matching_rules[0]}]` = '{search_term}'"
        select_statement = (
            "named_struct("
            + ", ".join(
                [
                    f"'{rule_name}', named_struct('column_name', '[{rule_name}]', 'value', `[{rule_name}]`)"
                    for rule_name in search_matching_rules
                ]
            )
            + ") AS search_result"
        )

        where_statement = f"WHERE {sql_filter}"

        return self._msql(
            f"SELECT {select_statement}, to_json(struct(*)) AS row_content FROM {from_tables} {where_statement}",
            min_score=min_score,
        )

    def select_by_classes(
        self,
        from_tables: str = "*.*.*",
        by_classes: Optional[Union[List[str], str]] = None,
        min_score: Optional[float] = None,
    ):
        """Selects all columns in the lakehouse from tables that match ALL the given classes

        Args:
            from_tables (str, optional): The tables to be selected in format
                "catalog.schema.table", use "*" as a wildcard. Defaults to "*.*.*".
            by_classes (Union[List[str], str], optional): The classes to be used to
                search for columns. Defaults to None.
            min_score (float, optional): Defines the classification score or frequency
                threshold for columns to be considered during the scan. Defaults to None
                which means that all columns where at least one record matched the
                respective rule during the scan will be included. Has to be either None
                or between 0 and 1.

        Raises:
            ValueError: If the by_classes type is not valid

        Returns:
            DataFrame: A dataframe containing the UNION ALL results of the select"""

        Msql.validate_from_components(from_tables)

        if isinstance(by_classes, str):
            by_classes = [by_classes]
        elif isinstance(by_classes, list) and all(isinstance(elem, str) for elem in by_classes):
            by_classes = by_classes
        else:
            raise ValueError(
                f"The provided by_classes {by_classes} have the wrong type. Please provide"
                f" either a str or List[str]."
            )

        from_statement = (
            "named_struct("
            + ", ".join(
                [
                    f"'{class_name}', named_struct('column_name', '[{class_name}]', 'value', `[{class_name}]`)"
                    for class_name in by_classes
                ]
            )
            + ") AS classified_columns"
        )

        return self._msql(
            f"SELECT {from_statement}, to_json(struct(*)) AS row_content FROM {from_tables}", min_score=min_score
        )

    def delete_by_class(
        self,
        from_tables="*.*.*",
        by_class: str = None,
        values: Optional[Union[List[str], str]] = None,
        yes_i_am_sure: bool = False,
        min_score: Optional[float] = None,
    ):
        """Deletes all rows in the lakehouse that match any of the provided values in a column classified with the given class

        Args:
            from_tables (str, optional): The tables to delete from in format
                "catalog.schema.table", use "*" as a wildcard. Defaults to "*.*.*".
            by_class (str, optional): The class to be used to search for columns.
                Defaults to None.
            values (Union[List[str], str], optional): The values to be deleted.
                Defaults to None.
            yes_i_am_sure (bool, optional): Whether you are sure that you want to delete
                the data. If False prints the SQL statements instead of executing them. Defaults to False.
            min_score (float, optional): Defines the classification score or frequency
                threshold for columns to be considered during the scan. Defaults to None
                which means that all columns where at least one record matched the
                respective rule during the scan will be included. Has to be either None
                or between 0 and 1.

        Raises:
            ValueError: If the from_tables is not valid
            ValueError: If the by_class is not valid
            ValueError: If the values is not valid
        """

        Msql.validate_from_components(from_tables)

        if (by_class is None) or (not isinstance(by_class, str)):
            raise ValueError(f"Please provide a class to identify the columns to be matched on the provided values.")

        if values is None:
            raise ValueError(
                f"Please specify the values to be deleted. You can either provide a list of values or a single value."
            )
        elif isinstance(values, str):
            value_string = f"'{values}'"
        elif isinstance(values, list) and all(isinstance(elem, str) for elem in values):
            value_string = "'" + "', '".join(values) + "'"
        else:
            raise ValueError(
                f"The provided values {values} have the wrong type. Please provide" f" either a str or List[str]."
            )

        if not yes_i_am_sure:
            self.logger.friendly(
                f"Please confirm that you want to delete the following values from the table {from_tables} using the class {by_class}: {values}"
            )
            self.logger.friendly(
                f"If you are sure, please run the same command again but set the parameter yes_i_am_sure to True."
            )

        delete_result = self._msql(
            f"DELETE FROM {from_tables} WHERE `[{by_class}]` IN ({value_string})",
            what_if=(not yes_i_am_sure),
            min_score=min_score,
        )

        if delete_result is not None:
            delete_result = delete_result.toPandas()
            self.logger.friendlyHTML(f"<p>The affcted tables are</p>{delete_result.to_html()}")

    def _msql(self, msql: str, what_if: bool = False, min_score: Optional[float] = None):
        self.logger.debug(f"Executing sql template: {msql}")

        msql_builder = Msql(msql)

        # check if classification is available
        # Check for more specific exception
        classification_result_pdf = self._scan_result.get_classes(min_score)
        sql_rows = msql_builder.build(classification_result_pdf)

        if what_if:
            self.logger.friendly(f"SQL that would be executed:")

            for sql_row in sql_rows:
                self.logger.friendly(sql_row.sql)

            return None
        else:
            self.logger.debug(f"Executing SQL:\n{sql_rows}")
            return msql_builder.execute_sql_rows(sql_rows, self.spark)



"""
This module contains common utility functions used throughout the package
"""
import sys


def strip_margin(text):
    """Trim margin in strings

    Args:
        text (string): The string

    Returns:
        string: The trimmed string

    """
    if not text:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = text.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return "\n".join(trimmed)


def format_regex(expression):
    return expression.replace("\\", r"\\")



"""
This file configures the Python package with entrypoints used for future runs on Databricks.

Please follow the `entry_points` documentation for more details on how to configure the entrypoint:
* https://setuptools.pypa.io/en/latest/userguide/entry_point.html
"""
import re
from os import path
from setuptools import find_packages, setup


DESCRIPTION = "DiscoverX - Map and Search your Lakehouse"

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

# import version
with open(path.join(this_directory, "discoverx/version.py"), encoding="utf-8") as f:
    version_line = f.read()
VERSION_PATTERN = r"^__version__ = ['\"]([^'\"]*)['\"]"
parsed_version = re.search(VERSION_PATTERN, version_line, re.M)
if parsed_version:
    version_string = parsed_version.group(1)
else:
    raise RuntimeError("Unable to find version string in discoverx/version.py")

PACKAGE_REQUIREMENTS = ["pyyaml"]

# packages for local development and unit testing
# please note that these packages are already available in DBR, there is no need to install them on DBR.
LOCAL_REQUIREMENTS = [
    "pyspark>=3.3.0",
    "delta-spark>=2.2.0",
    "pandas<2.0.0",  # From 2.0.0 onwards, pandas does not support iteritems() anymore, spark.createDataFrame will fail
    "numpy<1.24",  # From 1.24 onwards, module 'numpy' has no attribute 'bool'.
]

TEST_REQUIREMENTS = [
    # development & testing tools
    "pytest",
    "pylint",
    "black",
    "coverage[toml]",
    "pytest-cov",
    "dbx>=0.7,<0.9",
]

setup(
    name="dbl_discoverx",
    packages=find_packages(exclude=["tests", "tests.*"]),
    setup_requires=["setuptools", "wheel"],
    install_requires=PACKAGE_REQUIREMENTS,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
    extras_require={"local": LOCAL_REQUIREMENTS, "test": TEST_REQUIREMENTS},
    version=version_string,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://databricks.com/learn/labs",
    author="Erni Durdevic, David Tempelmann",
    author_email="labs@databricks.com",
    license_files=("LICENSE",),
)



import logging
import re


class Logging:
    def friendly(self, message):
        print(re.sub("<[^<]+?>", "", message))
        logging.info(message)

    def friendlyHTML(self, message):
        try:
            from dbruntime.display import displayHTML  # pylint: disable=import-error

            displayHTML(message)
        except:
            # Strip HTML classes
            print(re.sub("<[^<]+?>", "", message))
        logging.info(message)

    def info(self, message):
        print(message)
        logging.info(message)

    def debug(self, message):
        logging.debug(message)

    def error(self, message):
        print(message)
        logging.error(message)
